/***********************************************************************/
/*                                                                     */
/*                                                                     */
/* Program:                                                            */
/* --------                                                            */
/*                                                                     */
/* nbody.c                                                             */
/*                                                                     */
/*                                                                     */
/* Version:                                                            */
/* --------                                                            */
/*                                                                     */
/* First version: February 1998                                        */
/* Latest update: 2008-07-27                                           */
/*                                                                     */
/*                                                                     */
/* Author of the program:                                              */
/* ----------------------                                              */
/*                                                                     */
/* Stefan Spaennare, Lund, Sweden                                      */
/*                                                                     */
/* E-mail: stefan@spaennare.se                                         */
/*                                                                     */
/* Web: http://www.spaennare.se/index.html                             */
/* Web: http://www.spaennare.se/ssnbody.html                           */
/*                                                                     */
/* If you have problems with the program please let me know!           */
/*                                                                     */
/*                                                                     */
/* Documentation:                                                      */
/* --------------                                                      */
/*                                                                     */
/* For a detailed description of the ssnbody nbody package see the     */
/* files ssnbody.pdf and ssnbody.ps on the ssnbody web page.           */
/*                                                                     */
/*                                                                     */
/* Program history:                                                    */
/* ----------------                                                    */
/*                                                                     */
/* July 2008:                                                          */
/*                                                                     */
/* Extended mode output added.                                         */
/*                                                                     */
/* September 2006:                                                     */
/*                                                                     */
/* Decimals added to the constant pi to work with long double          */
/* precision. Very minor changes to the text output from the           */
/* program.                                                            */
/*                                                                     */
/* February 2000:                                                      */
/*                                                                     */
/* Optional calculation of relativistic (PPN) acceleration added.      */
/* Runge Kutta of 5th and 8th order and Bulirsch Stoer integration     */
/* added. Option for long double calculation added.                    */
/*                                                                     */
/* February 1998:                                                      */
/*                                                                     */
/* Simple non relativistic (Newtonian) nbody program with 4th order    */
/* Runge Kutta integration.                                            */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define LDOUBLE 0

#if LDOUBLE
#define double long double
#define sqrt sqrtl
#define floor floorl
#define fabs fabsl
#endif 

#define sqr(x) ((x)*(x))
#define cube(x) ((x)*(x)*(x))
#define min(x,y) (((x) < (y)) ? (x) : (y))
#define max(x,y) (((x) > (y)) ? (x) : (y))

double timetic=(double)(CLOCKS_PER_SEC);

double pi=3.1415926535897932384626433;


/*** Definitions for random (ran2) function.       ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (ran2).                      ***/

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)


/*** Random function.                              ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (ran2).                      ***/

double ran2(idum)
long *idum;
{
   int j;
   long k;
   static long idum2=123456789;
   static long iy=0; 
   static long iv[NTAB];
   double temp;

   if (*idum <= 0) {
      if (-(*idum) < 1) {
         *idum=1;
      }
      else {
         *idum=-(*idum);
      } /* if */
      idum2=(*idum);
      for (j=NTAB+7; j>=0; j--) {
         k=(*idum)/IQ1;
	 *idum=IA1*(*idum-k*IQ1)-k*IR1;
	 if (*idum < 0) {
	    *idum += IM1;
	 } /* if */
	 if (j < NTAB) {
	    iv[j]=*idum;
	 } /* if */
      } /* for j */
      iy=iv[0];
   } /* if */
   k=(*idum)/IQ1;
   *idum=IA1*(*idum-k*IQ1)-k*IR1;
   if (*idum < 0) {
      *idum += IM1;
   } /* if */
   k=idum2/IQ2;
   idum2=IA2*(idum2-k*IQ2)-k*IR2;
   if (idum2 < 0) {
      idum2 += IM2;
   } /* if */
   j=iy/NDIV;
   iy=iv[j]-idum2;
   iv[j] = *idum;
   if (iy < 1) {
      iy += IMM1;
   } /* if */
   temp=AM*(double)(iy);
   if (temp > RNMX) {
      return(RNMX);
   }
   else {
      return(temp);
   } /* if */

} /* ran2 */


/*** Setting up arrays for 8-order Runge-Kutta.           ***/
/*** From rksuite.f. Translated to C by Stefan Spaennare. ***/
/*** Reference: http://www.netlib.org/ode/index.html      ***/

void rk8_init(ptr,a,bhat,b,c,nstage)
long *ptr;
double **a;
double *bhat;
double *b;
double *c;
long nstage;
{

    long i,j;

    for (i=1; i<=nstage; i++) {
       ptr[i]=0;
       bhat[i]=0.0;
       b[i]=0.0;
       c[i]=0.0;
    } /* for i */

    for (i=1; i<=nstage; i++) {
       for (j=1; j<=nstage; j++) {
          a[i][j]=0.0;
       } /* for j */
    } /* for i */


      ptr[1] = 0;
      ptr[2] = 1;
      ptr[3] = 2;
      ptr[4] = 1;
      ptr[5] = 3;
      ptr[6] = 2;
      ptr[7] = 4;
      ptr[8] = 5;
      ptr[9] = 6;
      ptr[10] = 7;
      ptr[11] = 8;
      ptr[12] = 9;
      ptr[13] = 1;



      a[2][1] = 5.55555555555555555555555555556e-2;

      a[3][1] = 2.08333333333333333333333333333e-2;
      a[3][2] = 6.25e-2;

      a[4][1] = 3.125e-2;
      a[4][2] = 0.e0;
      a[4][3] = 9.375e-2;

      a[5][1] = 3.125e-1;
      a[5][2] = 0.e0;
      a[5][3] = -1.171875e0;
      a[5][4] = 1.171875e0;

      a[6][1] = 3.75e-2;
      a[6][2] = 0.e0;
      a[6][3] = 0.e0;
      a[6][4] = 1.875e-1;
      a[6][5] = 1.5e-1;

      a[7][1] = 4.79101371111111111111111111111e-2;
      a[7][2] = 0.e0;
      a[7][3] = 0.0e0;
      a[7][4] = 1.12248712777777777777777777778e-1;
      a[7][5] = -2.55056737777777777777777777778e-2;
      a[7][6] = 1.28468238888888888888888888889e-2;

      a[8][1] = 1.6917989787292281181431107136e-2;
      a[8][2] = 0.e0;
      a[8][3] = 0.e0;
      a[8][4] = 3.87848278486043169526545744159e-1;
      a[8][5] = 3.59773698515003278967008896348e-2;
      a[8][6] = 1.96970214215666060156715256072e-1;
      a[8][7] = -1.72713852340501838761392997002e-1;

      a[9][1] = 6.90957533591923006485645489846e-2;
      a[9][2] = 0.e0;
      a[9][3] = 0.e0;
      a[9][4] = -6.34247976728854151882807874972e-1;
      a[9][5] = -1.61197575224604080366876923982e-1;
      a[9][6] = 1.38650309458825255419866950133e-1;
      a[9][7] = 9.4092861403575626972423968413e-1;
      a[9][8] = 2.11636326481943981855372117132e-1;

      a[10][1] = 1.83556996839045385489806023537e-1;
      a[10][2] = 0.e0;
      a[10][3] = 0.e0;
      a[10][4] = -2.46876808431559245274431575997e0;
      a[10][5] = -2.91286887816300456388002572804e-1;
      a[10][6] = -2.6473020233117375688439799466e-2;
      a[10][7] = 2.84783876419280044916451825422e0;
      a[10][8] = 2.81387331469849792539403641827e-1;
      a[10][9] = 1.23744899863314657627030212664e-1;

      a[11][1] = -1.21542481739588805916051052503e0;
      a[11][2] = 0.e0;
      a[11][3] = 0.e0;
      a[11][4] = 1.66726086659457724322804132886e1;
      a[11][5] = 9.15741828416817960595718650451e-1;
      a[11][6] = -6.05660580435747094755450554309e0;
      a[11][7] = -1.60035735941561781118417064101e1;
      a[11][8] = 1.4849303086297662557545391898e1;
      a[11][9] = -1.33715757352898493182930413962e1;
      a[11][10] = 5.13418264817963793317325361166e0;

      a[12][1] = 2.58860916438264283815730932232e-1;
      a[12][2] = 0.e0;
      a[12][3] = 0.e0;
      a[12][4] = -4.77448578548920511231011750971e0;
      a[12][5] = -4.3509301377703250944070041181e-1;
      a[12][6] = -3.04948333207224150956051286631e0;
      a[12][7] = 5.57792003993609911742367663447e0;
      a[12][8] = 6.15583158986104009733868912669e0;
      a[12][9] = -5.06210458673693837007740643391e0;
      a[12][10] = 2.19392617318067906127491429047e0;
      a[12][11] = 1.34627998659334941535726237887e-1;

      a[13][1] = 8.22427599626507477963168204773e-1;
      a[13][2] = 0.e0;
      a[13][3] = 0.e0;
      a[13][4] = -1.16586732572776642839765530355e1;
      a[13][5] = -7.57622116690936195881116154088e-1;
      a[13][6] = 7.13973588159581527978269282765e-1;
      a[13][7] = 1.20757749868900567395661704486e1;
      a[13][8] = -2.12765911392040265639082085897e0;
      a[13][9] = 1.99016620704895541832807169835e0;
      a[13][10] = -2.34286471544040292660294691857e-1;
      a[13][11] = 1.7589857770794226507310510589e-1;
      a[13][12] = 0.e0;

      bhat[1] = 4.17474911415302462220859284685e-2;
      bhat[2] = 0.e0;
      bhat[3] = 0.e0;
      bhat[4] = 0.e0;
      bhat[5] = 0.e0;
      bhat[6] = -5.54523286112393089615218946547e-2;
      bhat[7] = 2.39312807201180097046747354249e-1;
      bhat[8] = 7.0351066940344302305804641089e-1;
      bhat[9] = -7.59759613814460929884487677085e-1;
      bhat[10] = 6.60563030922286341461378594838e-1;
      bhat[11] = 1.58187482510123335529614838601e-1;
      bhat[12] = -2.38109538752862804471863555306e-1;
      bhat[13] = 2.5e-1;

      b[1] = 2.9553213676353496981964883112e-2;
      b[2] = 0.e0;
      b[3] = 0.e0;
      b[4] = 0.e0;
      b[5] = 0.e0;
      b[6] = -8.28606276487797039766805612689e-1;
      b[7] = 3.11240900051118327929913751627e-1;
      b[8] = 2.46734519059988698196468570407e0;
      b[9] = -2.54694165184190873912738007542e0;
      b[10] = 1.44354858367677524030187495069e0;
      b[11] = 7.94155958811272872713019541622e-2;
      b[12] = 4.44444444444444444444444444445e-2;
      b[13] = 0.e0;

      c[1] = 0.e0;
      c[2] = 5.55555555555555555555555555556e-2;
      c[3] = 8.33333333333333333333333333334e-2;
      c[4] = 1.25e-1;
      c[5] = 3.125e-1;
      c[6] = 3.75e-1;
      c[7] = 1.475e-1;
      c[8] = 4.65e-1;
      c[9] = 5.64865451382259575398358501426e-1;
      c[10] = 6.5e-1;
      c[11] = 9.24656277640504446745013574318e-1;
      c[12] = 1.e0;
      c[13] = 1.e0; 


} /* rk8_init */


/*** Simple procedure for setting up random positions       ***/
/*** and velocities for the bodies. Spherical distribution. ***/          

void init_random(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,n,rmax,vmax,mmax,mmin,idum)
double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
long n;
double rmax,vmax,mmax,mmin;
long *idum;
{

  long i;
  double w,w2,rr;

  for (i=1; i<=n; i++) {

     w=2.0*pi*ran2(idum);
     w2=pi*ran2(idum);
     rr=rmax*sqrt(ran2(idum));

     xx1[i]=rr*sin(w2)*cos(w);
     yy1[i]=rr*sin(w2)*sin(w);
     zz1[i]=rr*cos(w2);

     vxx1[i]=2.0*vmax*(ran2(idum)-0.5);
     vyy1[i]=2.0*vmax*(ran2(idum)-0.5);
     vzz1[i]=2.0*vmax*(ran2(idum)-0.5);

     mm0[i]=(mmax-mmin)*ran2(idum)+mmin;

  } /* for i */

} /* init_random */


/*** Calculating position and velocity for system barycentrum. ***/
/*** Non relativistic (Newtonian) method.                      ***/

void barycentrum_newt(rbar,vbar,mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,n,g)
double *rbar;
double *vbar;
double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
long n;
double g;
{
   long i;

   double xs,ys,zs,usum,usumi,ui;
   double vxs,vys,vzs;

   xs=0.0;
   ys=0.0;
   zs=0.0;

   vxs=0.0;
   vys=0.0;
   vzs=0.0;

   usum=0.0;

   for (i=1; i<=n; i++) {

      ui=g*mm0[i];

      usum=usum+ui;

      xs=xs+xx1[i]*ui;
      ys=ys+yy1[i]*ui;
      zs=zs+zz1[i]*ui;

      vxs=vxs+vxx1[i]*ui;
      vys=vys+vyy1[i]*ui;
      vzs=vzs+vzz1[i]*ui;

   } /* for i */

   usumi=1.0/usum;

   rbar[1]=xs*usumi;
   rbar[2]=ys*usumi;
   rbar[3]=zs*usumi;

   vbar[1]=vxs*usumi;
   vbar[2]=vys*usumi;
   vbar[3]=vzs*usumi;

} /* barycentrum_newt */


/*** Calculates the non relativistic (Newtonian) accelerations for ***/
/*** the bodies.                                                   ***/
/*** Reference: Lennart Lindegren, Lund Observatory, private       ***/
/*** communication.                                                ***/

void compacc_newt(mm0,xx1,yy1,zz1,axx1,ayy1,azz1,n,g,mfr,rmin)
double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *axx1;
double *ayy1;
double *azz1;
long n;
double g,mfr;
double *rmin;
{

   long i,j;

   double xi,yi,zi,ax,ay,az,r,r2,r3g;
   double dx,dy,dz,mj,mi,mfr2,rmin2;

   for (i=1; i<=n; i++) {
      axx1[i]=0.0;
      ayy1[i]=0.0;
      azz1[i]=0.0;
   } /* for i */   

   mfr2=mfr*mfr;

   rmin2=1.0e30;

   for (i=2; i<=n; i++) {   

      mi=mm0[i];
      xi=xx1[i];
      yi=yy1[i];
      zi=zz1[i];

      for (j=1; j<i; j++) {

            mj=mm0[j];

            dx=xi-xx1[j];
	    dy=yi-yy1[j];
	    dz=zi-zz1[j];

            r2=dx*dx+dy*dy+dz*dz;

           if (r2 < rmin2) {
	      rmin2=r2;
	   } /* if */

            if (r2 < mfr2) { 
               r2=mfr2;
            } /* if */

	    r=sqrt(r2);

	    r3g=g/(r*r2);
            
	    ax=dx*r3g;
	    ay=dy*r3g;
	    az=dz*r3g;

	    axx1[i]=axx1[i]-ax*mj;
	    ayy1[i]=ayy1[i]-ay*mj;
	    azz1[i]=azz1[i]-az*mj;

	    axx1[j]=axx1[j]+ax*mi;
	    ayy1[j]=ayy1[j]+ay*mi;
	    azz1[j]=azz1[j]+az*mi;

      } /* for j */

   } /* for i */

   *rmin=sqrt(rmin2);

} /* compacc_newt */


/*** Calculates non relativistic (Newtonian) barycentrum if relmode = 0  ***/
/*** and relativistic (PPN) barycentrum if relmode = 1.                  ***/
/*** Reference (relativistic method):                                    ***/
/*** Newhall X. X., Standish E. M. Jr., Williams J. G.: 1983,            ***/
/*** Astron. Astrophys. 125, 150-167, "DE 102: a nummerically integrated ***/
/*** ephemeris of the Moon and planets spanning forty-four centuries".   ***/

void barycentrum(rbar,vbar,mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,n,g,mfr,cc,relmode)
double *rbar;
double *vbar;
double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
long n;
double g,mfr,cc;
long relmode;
{
   long i,j;

   double xs,ys,zs;
   double cci22,vi2,mfr2;
   double xi,yi,zi,dx,dy,dz;
   double vxi,vyi,vzi,usum,usumi;
   double psum,r,r2,ui,uj,mi;
   double vxs,vys,vzs;

if (relmode != 1) {

   barycentrum_newt(rbar,vbar,mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,n,g);

}
else {

   mfr2=mfr*mfr;

   cci22=1.0/(2.0*cc*cc);

   xs=0.0;
   ys=0.0;
   zs=0.0;

   vxs=0.0;
   vys=0.0;
   vzs=0.0;

   usum=0.0;

   for (i=1; i<=n; i++) {

      mi=mm0[i];

      xi=xx1[i];
      yi=yy1[i];
      zi=zz1[i];

      vxi=vxx1[i];
      vyi=vyy1[i];
      vzi=vzz1[i];

      vi2=vxi*vxi+vyi*vyi+vzi*vzi;

      psum=0.0;

      for (j=1; j<=n; j++) {
      
         if (j != i) {
	    
	    uj=g*mm0[j];

	    dx=xi-xx1[j];
	    dy=yi-yy1[j];
	    dz=zi-zz1[j];

	    r2=dx*dx+dy*dy+dz*dz;

            if (r2 < mfr2) { 
               r2=mfr2;
            } /* if */

	    r=sqrt(r2);

	    psum=psum+uj/r;

         } /* if */

      } /* for j */

      ui=g*mi*(1.0+cci22*(vi2-psum));

      usum=usum+ui;

      xs=xs+xi*ui;
      ys=ys+yi*ui;
      zs=zs+zi*ui;

      vxs=vxs+vxi*ui;
      vys=vys+vyi*ui;
      vzs=vzs+vzi*ui;


   } /* for i */

   usumi=1.0/usum;

   rbar[1]=xs*usumi;
   rbar[2]=ys*usumi;
   rbar[3]=zs*usumi;

   vbar[1]=vxs*usumi;
   vbar[2]=vys*usumi;
   vbar[3]=vzs*usumi;

} /* if */

} /* barycentrum */


/*** Calculates non relativistic (Newtonian) accelerations  for the      ***/
/*** bodies if relmode = 0  and relativistic (PPN) accelerations for     ***/
/*** the bodies if relmode = 1.                                          ***/
/*** Reference (relativistic method):                                    ***/
/*** Newhall X. X., Standish E. M. Jr., Williams J. G.: 1983,            ***/
/*** Astron. Astrophys. 125, 150-167, "DE 102: a nummerically integrated ***/
/*** ephemeris of the Moon and planets spanning forty-four centuries".   ***/

void compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
             axxn,ayyn,azzn,qq,n,g,mfr,cc,rmin,relmode)
double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long n;
double g,mfr,cc;
double *rmin;
long relmode;
{

   long i,j;

   double xi,yi,zi,r,r2;
   double dx,dy,dz,mj,mi,mfr2,rmin2;
   double cci2,vxi,vyi,vzi,beta,gamma;
   double c1,c2,c3,c4,c5,c6,c7,ri;
   double vi2,vj2,vxj,vyj,vzj;
   double xj,yj,zj,pp,ss;
   double anxj,anyj,anzj;
   double rr,d1a,d1b,d2;
   double bxi,byi,bzi;
   double aaa,bbb,vvx,vvy,vvz;
   double r3i,uj,r3g,temp1,temp2,temp3;

if (relmode != 1) {

   compacc_newt(mm0,xx1,yy1,zz1,axx1,ayy1,azz1,n,g,mfr,rmin);

} 
else {

   cci2=1.0/(cc*cc);

   beta=1.0;
   gamma=1.0;

   c1=2.0*(beta+gamma);
   c2=2.0*beta-1.0;
   c3=gamma;
   c4=1.0+gamma;
   c5=2.0*(1.0+gamma);
   c6=0.5*3.0;
   c7=0.5;

   d1a=2.0+2.0*gamma;
   d1b=1.0+2.0*gamma;
   d2=0.5*(3.0+4.0*gamma);

   mfr2=mfr*mfr;

   rmin2=1.0e30;

   for (i=1; i<=n; i++) {
      qq[i]=0.0;

      axxn[i]=0.0;
      ayyn[i]=0.0;
      azzn[i]=0.0;
   } /* for i */   

   for (i=1; i<=n; i++) {   

      mi=mm0[i];
      xi=xx1[i];
      yi=yy1[i];
      zi=zz1[i];

      for (j=1; j<=n; j++) {

         if (j != i) {

            uj=g*mm0[j];

            dx=xi-xx1[j];
	    dy=yi-yy1[j];
	    dz=zi-zz1[j];

            r2=dx*dx+dy*dy+dz*dz;

            if (r2 < rmin2) {
	       rmin2=r2;
	    } /* if */

            if (r2 < mfr2) { 
               r2=mfr2;
            } /* if */

	    r=sqrt(r2);

	    r3g=uj/(r*r2);

	    qq[i]=qq[i]+uj/r;

	    axxn[i]=axxn[i]-r3g*dx;
	    ayyn[i]=ayyn[i]-r3g*dy;
	    azzn[i]=azzn[i]-r3g*dz;

         } /* if */

      } /* for j */

   } /* for i */


   for (i=1; i<=n; i++) {
      axx1[i]=0.0;
      ayy1[i]=0.0;
      azz1[i]=0.0;
   } /* for i */   

   for (i=1; i<=n; i++) {   

      mi=mm0[i];
      xi=xx1[i];
      yi=yy1[i];
      zi=zz1[i];
      vxi=vxx1[i];
      vyi=vyy1[i];
      vzi=vzz1[i];

      vi2=vxi*vxi+vyi*vyi+vzi*vzi;

      for (j=1; j<=n; j++) {

         if (j != i) {

            mj=mm0[j];
	    xj=xx1[j];
	    yj=yy1[j];
	    zj=zz1[j];
	    vxj=vxx1[j];
	    vyj=vyy1[j];
	    vzj=vzz1[j];

	    uj=g*mj;

	    vj2=vxj*vxj+vyj*vyj+vzj*vzj;

	    anxj=axxn[j];
	    anyj=ayyn[j];
	    anzj=azzn[j];

            dx=xi-xj;
	    dy=yi-yj;
	    dz=zi-zj;

            r2=dx*dx+dy*dy+dz*dz;

            if (r2 < mfr2) { 
               r2=mfr2;
            } /* if */

	    r=sqrt(r2);

	    ri=1.0/r;

	    r3i=1.0/(r*r2);
            
	    rr=vxi*vxj+vyi*vyj+vzi*vzj;

	    pp=ri*(dx*vxj+dy*vyj+dz*vzj);

	    ss=-dx*anxj-dy*anyj-dz*anzj;

	    aaa=1.0+cci2*(-c1*qq[i]-c2*qq[j]+c3*vi2+c4*vj2-c5*rr-c6*pp*pp+c7*ss);

	    vvx=vxi-vxj;
	    vvy=vyi-vyj;
	    vvz=vzi-vzj;

	    bxi=d1a*vxi-d1b*vxj;
	    byi=d1a*vyi-d1b*vyj;
	    bzi=d1a*vzi-d1b*vzj;

	    bbb=dx*bxi+dy*byi+dz*bzi;

	    temp1=uj*r3i*aaa;
	    temp2=cci2*uj*r3i*bbb;
	    temp3=cci2*uj*ri*d2;

	    axx1[i]=axx1[i]-temp1*dx+temp2*vvx+temp3*anxj;
	    ayy1[i]=ayy1[i]-temp1*dy+temp2*vvy+temp3*anyj;
	    azz1[i]=azz1[i]-temp1*dz+temp2*vvz+temp3*anzj;

         } /* if */

      } /* for j */

   } /* for i */

   *rmin=sqrt(rmin2);

} /* if */

} /* compacc */


/*** Calculates minimum distance between bodies. ***/

void comprmin(xx1,yy1,zz1,n,rmin)
double *xx1;
double *yy1;
double *zz1;
long n;
double *rmin;
{

   long i,j;

   double dx,dy,dz,r2,rmin2;

   rmin2=1.0e30;

   for (i=2; i<=n; i++) {   

      for (j=1; j<i; j++) {

            dx=xx1[i]-xx1[j];
	    dy=yy1[i]-yy1[j];
	    dz=zz1[i]-zz1[j];

            r2=dx*dx+dy*dy+dz*dz;

           if (r2 < rmin2) {
	      rmin2=r2;
	   } /* if */

      } /* for j */

   } /* for i */

   *rmin=sqrt(rmin2);

} /* comprmin */


/*** Calculates maximum velocity of bodies. ***/

void compvmax(vxx1,vyy1,vzz1,n,vmax)
double *vxx1;
double *vyy1;
double *vzz1;
long n;
double *vmax;
{
   long j;
   double vmax2,dvx,dvy,dvz,vvv;

   vmax2=0.0;

   for (j=1; j<=n; j++) {
     dvx=vxx1[j];
     dvy=vyy1[j];
     dvz=vzz1[j];
     vvv=dvx*dvx+dvy*dvy+dvz*dvz;
     if (vvv > vmax2) {
        vmax2=vvv;
     } /* if */
   } /* for j */

   *vmax=sqrt(vmax2);

} /* compvmax */


/*** Converts position (x,y,z) and velocity (vx,vy,vz) ***/
/*** to a singel vector y.                             ***/

void rv2y(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,n)
double *y;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
long n;

{
    long k,i;

    k=1;

    for (i=1; i<=n; i++) {
       y[k]=xx1[i];
       k++;
       y[k]=yy1[i];
       k++;
       y[k]=zz1[i];
       k++;
    } /* for i */

    for (i=1; i<=n; i++) {
       y[k]=vxx1[i];
       k++;
       y[k]=vyy1[i];
       k++;
       y[k]=vzz1[i];
       k++;
    } /* for i */

} /* rv2y */


/*** Converts a single vector y to positions (x,y,z) and ***/
/*** velocities (vx,vy,vz).                              ***/

void y2rv(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,n)
double *y;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
long n;

{
    long k,i;

    k=1;

    for (i=1; i<=n; i++) {
       xx1[i]=y[k];
       k++;
       yy1[i]=y[k];
       k++;
       zz1[i]=y[k];
       k++;
    } /* for i */

    for (i=1; i<=n; i++) {
       vxx1[i]=y[k];
       k++;
       vyy1[i]=y[k];
       k++;
       vzz1[i]=y[k];
       k++;
    } /* for i */

} /* y2rv */


/*** Runge-Kutta 4-order.                              ***/
/*** Reference: Lennart Lindegren, Lund Observatory,   ***/
/*** private communication. The algorithm can also     ***/
/*** be found elsewhere on Internet.                   ***/


void rk4(h,
         mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
         axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode)

double h;	    

double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long npart;
double g,mfr,cc;
double *rrmin;
long relmode;
{

   long j,n,n6;

   double ss;

   double *k1,*k2,*k3,*k4;

   double *y1,*y11;

   n=npart;
   n6=6*n;

  k1=(double *)calloc(n6+1,sizeof(double));
  k2=(double *)calloc(n6+1,sizeof(double));
  k3=(double *)calloc(n6+1,sizeof(double));
  k4=(double *)calloc(n6+1,sizeof(double));

  y1=(double *)calloc(n6+1,sizeof(double));
  y11=(double *)calloc(n6+1,sizeof(double));


   ss=1.0/6.0;


   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y1,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);
   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      k1[j]=y11[j]*h;
      y11[j]=y1[j]+0.5*k1[j];

   } /* for j */
 
   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      k2[j]=y11[j]*h;
      y11[j]=y1[j]+0.5*k2[j];

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      k3[j]=y11[j]*h;
      y11[j]=y1[j]+k3[j];

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      k4[j]=y11[j]*h;

   } /* for j */

   for (j=1; j<=n6; j++) {

      y11[j]=y1[j]+ss*(k1[j]+2.0*(k2[j]+k3[j])+k4[j]);

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   free(k1);
   free(k2);
   free(k3);
   free(k4);

   free(y1);
   free(y11);

} /* rk4 */


/*** Runge-Kutta 5-order.                          ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (rkck).                      ***/

void rk5(h,xerr1,yerr1,zerr1,
         mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
         axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode)

double h;	    
double *xerr1;
double *yerr1;
double *zerr1;

double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long npart;
double g,mfr,cc;
double *rrmin;
long relmode;
{

   long j,k,n,n6;

   double temp;
   double a2,a3,a4,a5,a6;
   double b21,b31,b32,b41,b42,b43,b51,b52,b53,b54;
   double b61,b62,b63,b64,b65;
   double c1,c3,c4,c6;
   double dc1,dc3,dc4,dc5,dc6;

   double *ak2,*ak3,*ak4,*ak5,*ak6;

   double *y1,*y11,*y2;

   a2=0.2; a3=0.3; a4=0.6; a5=1.0; a6=0.875;

   b21=0.2; b31=3.0/40.0; b32=9.0/40.0; 
   
   b41=0.3; b42=-0.9; b43=1.2;

   b51=-11.0/54.0; b52=2.5; b53=-70.0/27.0; b54=35.0/27.0;

   b61=1631.0/55296.0; b62=175.0/512.0; b63=575.0/13824.0;
   b64=44275.0/110592.0; b65=253.0/4096.0;

   c1=37.0/378.0; c3=250.0/621.0; c4=125.0/594.0; c6=512.0/1771.0;

   dc1=c1-2825.0/27648.0;
   dc3=c3-18575.0/48384.0;
   dc4=c4-13525.0/55296.0;
   dc5=-277.0/14336.0;
   dc6=c6-0.25;

   n=npart;
   n6=6*n;

  y1=(double *)calloc(n6+1,sizeof(double));
  y11=(double *)calloc(n6+1,sizeof(double));
  y2=(double *)calloc(n6+1,sizeof(double));


  ak2=(double *)calloc(n6+1,sizeof(double));
  ak3=(double *)calloc(n6+1,sizeof(double));
  ak4=(double *)calloc(n6+1,sizeof(double));
  ak5=(double *)calloc(n6+1,sizeof(double));
  ak6=(double *)calloc(n6+1,sizeof(double));


   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y1,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);
   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);
   rv2y(y2,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      temp=b21*y2[j];
      y11[j]=y1[j]+h*temp;

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      ak2[j]=y11[j];
      temp=b31*y2[j]+b32*ak2[j];
      y11[j]=y1[j]+h*temp;

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      ak3[j]=y11[j];
      temp=b41*y2[j]+b42*ak2[j]+b43*ak3[j];
      y11[j]=y1[j]+h*temp;

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);



   for (j=1; j<=n6; j++) {

      ak4[j]=y11[j];
      temp=b51*y2[j]+b52*ak2[j]+b53*ak3[j]+b54*ak4[j];
      y11[j]=y1[j]+h*temp;

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);


   for (j=1; j<=n6; j++) {

      ak5[j]=y11[j];
      temp=b61*y2[j]+b62*ak2[j]+b63*ak3[j]+b64*ak4[j]+b65*ak5[j];
      y11[j]=y1[j]+h*temp;

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,n,g,mfr,cc,rrmin,relmode);

   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,n);

   for (j=1; j<=n6; j++) {

      ak6[j]=y11[j];
      temp=c1*y2[j]+c3*ak3[j]+c4*ak4[j]+c6*ak6[j];
      y11[j]=y1[j]+h*temp;
      y2[j]=h*(dc1*y2[j]+dc3*ak3[j]+dc4*ak4[j]+dc5*ak5[j]+dc6*ak6[j]);

   } /* for j */

   y2rv(y11,xx1,yy1,zz1,vxx1,vyy1,vzz1,n);

   k=1;
   for (j=1; j<=n; j++) {

      xerr1[j]=xerr1[j]+y2[k];
      k++;
      yerr1[j]=yerr1[j]+y2[k];
      k++;
      zerr1[j]=zerr1[j]+y2[k];
      k++;

   } /* for j */

   free(ak2);
   free(ak3);
   free(ak4);
   free(ak5);
   free(ak6);

   free(y1);
   free(y11);
   free(y2);

} /* rk5 */


/*** Runge-Kutta 8-order.                                 ***/
/*** From rksuite.f. Translated to C by Stefan Spaennare. ***/
/*** Reference: http://www.netlib.org/ode/index.html      ***/

void rk8(ptr,a,bhat,b,c,h,y,stages,nstage,xerr1,yerr1,zerr1,
            mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
            axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode)
long *ptr;
double **a;
double *bhat;
double *b;
double *c;

double h;
double *y;
double **stages;
long nstage;

double *xerr1;
double *yerr1;
double *zerr1;

double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long npart;
double g,mfr,cc;
double *rrmin;
long relmode;

{


   long i,j,k,kk,n6,fsal;

   double *y1,*y11;

   double *ee;

   n6=6*npart;

   y1=(double *)calloc(n6+1,sizeof(double));
   y11=(double *)calloc(n6+1,sizeof(double));

   ee=(double *)calloc(n6+1,sizeof(double));

   fsal=1;

   compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
           axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);


   rv2y(y1,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);
   rv2y(y11,vxx1,vyy1,vzz1,axx1,ayy1,azz1,npart);

   for (k=1; k<=n6; k++) {

       stages[k][ptr[1]+1]=y11[k];

  } /* for k */    


  for (i=2; i<=nstage; i++) {

      for (j=1; j<=(i-1); j++) {

         if (j == 1) {

	    for (k=1; k<=n6; k++) {

	       ee[k]=a[i][1]*y11[k];

            } /* for j */

	}
	else {

	   for (k=1; k<=n6; k++) {

	      ee[k]=ee[k]+a[i][j]*stages[k][ptr[j]+1]; 

           } /* for k */

       } /* if */

    } /* for j */

       
       for (k=1; k<=n6; k++) {

           y[k]=y1[k]+h*ee[k];
           
       } /* for k */	   

     
      y2rv(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);
      

      compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
              axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);

      rv2y(y,vxx1,vyy1,vzz1,axx1,ayy1,azz1,npart);

       for (k=1; k<=n6; k++) {
           
	   stages[k][ptr[i]+1]=y[k];

       } /* for k */	   

   } /* for i */    


      if (fsal == 1) {
         for (i=1; i<=nstage; i++) {

	    if (i == 1) {
	       for (k=1; k<=n6; k++) {

	          ee[k]=bhat[1]*y11[k];
         
	      } /* for k */
          }
	  else {
	       for (k=1; k<=n6; k++) {
	          
		  ee[k]=ee[k]+bhat[i]*stages[k][ptr[i]+1];

              } /* for k */

	    } /* if */
	  
	 } /* for i */

	 for (k=1; k<=n6; k++) {

	    y[k]=y1[k]+h*ee[k];
	     
        } /* for k */

     }  /* if */


     for (i=1; i<=nstage; i++) {

        if (i == 1) {

	   for (k=1; k<=n6; k++) {
	      
	      ee[k]=ee[k]-b[1]*y11[k];
        
	   } /* for k */

       }
       else {
	   for (k=1; k<=n6; k++) {
	      
	      ee[k]=ee[k]-b[i]*stages[k][ptr[i]+1];
        
	   } /* for k */

       } /* if */
    } /* for i */

      kk=1;
      for (k=1; k<=npart; k++) {
        
	xerr1[k]=xerr1[k]+h*ee[kk];
	yerr1[k]=yerr1[k]+h*ee[kk+1];
	zerr1[k]=zerr1[k]+h*ee[kk+2];
	kk=kk+3;

     } /* for k */	
          
     y2rv(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

     free(ee);
     free(y1);
     free(y11);
        


} /* rk8 */


/*** Help routine for the Bulirsch-Stoer routine.  ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (mmid).                      ***/

void mmid(y, dydx, xs, htot, nstep, yout,
          mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
          axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode)

double *y;
double *dydx;
double xs;
double htot;
long nstep;
double *yout;

double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long npart;
double g,mfr,cc;
double *rrmin;
long relmode;

{

	 long i,n,n6;

	 double x,h,h2,swap;

	 double *ym,*yn;

	 n6=6*npart;

	 ym=(double *)calloc(n6+1,sizeof(double));
	 yn=(double *)calloc(n6+1,sizeof(double));

	 h = htot/(double)(nstep);

	 for (i=1;i<=n6;i++) {
	     ym[i] = y[i];
	     yn[i] = y[i] + h*dydx[i];
         } /* for i */

         x = xs + h;

	 y2rv(yn,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

         compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
                 axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);

	 rv2y(yout,vxx1,vyy1,vzz1,axx1,ayy1,azz1,npart);

         h2 = 2.0*h;

         for (n=2; n<=nstep; n++) {

            for (i=1;i<=n6;i++) {
	        swap = ym[i] + h2*yout[i];
                ym[i] = yn[i];
                yn[i] = swap;
             } /* for i */

	     x = x + h;

	     y2rv(yn,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

             compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
                     axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);

	     rv2y(yout,vxx1,vyy1,vzz1,axx1,ayy1,azz1,npart);


         } /* for n */

	 for (i=1;i<=n6;i++) {
	    yout[i] = 0.5*(ym[i] + yn[i] + h*yout[i]);
         } /* for i */

	 free(ym);
	 free(yn);

} /* mmid */


/*** Help routine for the Bulirsch-Stoer routine.  ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (pzextr).                    ***/

void pzextr(iest,xest,yest,yz,dy,d,x,npart)
long iest;
double xest;
double *yest;
double *yz;
double *dy;
double **d;
double *x;
long npart;
{
   long k1,j,n6;

   double q,f2,f1,delta;

   double *c;

   n6=6*npart;

   c=(double *)calloc(n6+1,sizeof(double));

   x[iest]=xest;

   for (j=1; j<=n6; j++) {

      dy[j]=yest[j];
      yz[j]=yest[j];

   } /* for j */

   if (iest == 1) {

      for (j=1; j<=n6; j++) {
         d[j][1]=yest[j];
      } /* for j */

   }
   else {

      for (j=1; j<=n6; j++) {
         c[j]=yest[j];
      } /* for j */
      
      for (k1=1; k1<iest; k1++) {

         delta=1.0/(x[iest-k1]-xest);

	 f1=xest*delta;
	 f2=x[iest-k1]*delta;

	 for (j=1; j<=n6; j++) {
	    q=d[j][k1];
	    d[j][k1]=dy[j];
	    delta=c[j]-q;
	    dy[j]=f1*delta;
	    c[j]=f2*delta;
	    yz[j]=yz[j]+dy[j];
         } /* for j */

      } /* for k1 */

      for (j=1; j<=n6; j++) {
         d[j][iest]=dy[j];
      } /* for j */

   } /* if */

   free(c);

} /* pzextr */


/*** Help routine for the Bulirsch-Stoer routine.  ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (rzextr).                    ***/

void rzextr(iest,xest,yest,yz,dy,d,x,npart)
long iest;
double xest;
double *yest;
double *yz;
double *dy;
double **d;
double *x;
long npart;

{
       long j,k,imax,nmax,ncol,kmax,n6;

       double yy,ddy,v,c1,b,b1;

       double *fx;

       n6=6*npart;

       imax=11;
       nmax=90;
       ncol=7;

       fx=(double *)calloc(iest+1,sizeof(double));

       kmax=8;

       x[iest] = xest;

       if (iest == 1) {

         for (j=1;j<=n6;j++) {

             yz[j] = yest[j];
             d[j][1] = yest[j];
             dy[j] = yest[j];

	 } /* for j */


        }
        else {

           for (k=1; k<iest; k++) {

               fx[k+1] = x[iest-k]/xest;

           } /* for k */

           for (j=1; j<=n6; j++) {

               yy = yest[j];
               v = d[j][1];
               c1 = yy;
               d[j][1] = yy;

               for (k=2; k<=iest; k++) {

                  b1 = fx[k]*v;
                  b = b1 - c1;

                  if (b != 0.0) {
                     b = (c1-v)/b;
                     ddy = c1*b;
                     c1 = b1*b;
                  }
                  else {
                     ddy = v;
                  } /* if */
                  
		  if (k != iest) {
                     v = d[j][k];
                  } /* if */

                  d[j][k] = ddy;
                  yy = yy + ddy;

              } /* for k */

              dy[j] = ddy;
              yz[j] = yy;

           } /* for j */

        } /* if */

	free(fx);


} /* rzextr */


/*** Bulirsch-Stoer routine.                       ***/
/*** Somewhat modified by Stefan Spaennare.        ***/
/*** Reference: Numerical Recipes in C second      ***/
/*** edition. Routine (bsstep).                    ***/

void bsstep(y,dydx,htry,eps,yscal,ddd,xxx,imax,
            mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
            axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode)

double *y;
double *dydx;
double htry;
double eps;
double *yscal;
double **ddd;
double *xxx;
long imax;

double *mm0;
double *xx1;
double *yy1;
double *zz1;
double *vxx1;
double *vyy1;
double *vzz1;
double *axx1;
double *ayy1;
double *azz1;
double *axxn;
double *ayyn;
double *azzn;
double *qq;
long npart;
double g,mfr,cc;
double *rrmin;
long relmode;


{

        long i,j,n6;

	double errmax,temp,h;
	double xsav,xest,ymax,ytemp;

	double *yerr;
	double *ysav;
	double *dysav;
	double *yseq;

	long *nseq;

	n6=6*npart;

	yerr=(double *)calloc(n6+1,sizeof(double));
	ysav=(double *)calloc(n6+1,sizeof(double));
	dysav=(double *)calloc(n6+1,sizeof(double));
	yseq=(double *)calloc(n6+1,sizeof(double));

	nseq=(long *)calloc(imax+1,sizeof(long));


	for (i=1; i<=imax; i++) {

	   nseq[i]=2*i;

        } /* for i */

	h = htry;
	xsav = 0.0;


        rv2y(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

        compacc(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
                 axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);

        rv2y(dydx,vxx1,vyy1,vzz1,axx1,ayy1,azz1,npart);


	for (i=1;i<=n6;i++) {
	    ysav[i] = y[i];
	    dysav[i] = dydx[i];
        } /* for i */


        for (i=1; i<=imax; i++) {

	   mmid(ysav,dysav,xsav,h,nseq[i],yseq,
                mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
                axxn,ayyn,azzn,qq,npart,g,mfr,cc,rrmin,relmode);


            xest = sqr(h/(double)(nseq[i]));

            rzextr(i,xest,yseq,y,yerr,ddd,xxx,npart);

            errmax = 0.0;
	    ymax=0.0;

            for (j=1; j<=n6; j++) {

		temp = fabs(yerr[j]*yscal[j]);
		errmax = max(errmax,temp);

		ytemp=fabs(y[j]);
		ymax=max(ymax,ytemp);
		
	    } /* for i */


            if (errmax < (eps*ymax)) {

                y2rv(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

	        free(yerr);
	        free(ysav);
	        free(dysav);
	        free(yseq);

		free(nseq);

		return;

	     } /*if */

         }  /* for i */

        y2rv(y,xx1,yy1,zz1,vxx1,vyy1,vzz1,npart);

        free(yerr);
        free(ysav);
        free(dysav);
        free(yseq);

	free(nseq);

        printf("bsstep failed\n");
        exit(0);

} /* bsstep */



int main(argc,argv) 
int argc;
char *argv[];
{
   
   FILE *outfile,*infile,*binfile;

   long n,i,j,idum,seed,flag1,flag2,flag3,inputflag;
   long printmode,barymode,n6,imax;
   long algmode,relmode,nstage;

   double rmax,vmax,mmax,mmin,g,mfr;
   double rrmin,vvmax,ddt;
   double tmax,tmax0,tp,tp2,tt,dtmin;
   double cc,dt,h,eps;
   double xbar,ybar,zbar,vxbar,vybar,vzbar;
   double bseps,dtsgn;

   double *yy;

   double *dydx;

   double *yscal;
    
   double *xxx; 

   double *qq;

   double *xerr1,*yerr1,*zerr1;

   double *xx1,*yy1,*zz1;
   double *vxx1,*vyy1,*vzz1;
   double *axx1,*ayy1,*azz1;

   double *axxn,*ayyn,*azzn;

   double *mm0;

   double **stages;
   double *mpstages;

   double **aaa;
   double *mpaaa;

   double *ccc;

   double *bbb;

   double *bhat;

   double **ddd;
   double *mpddd;

   double rbar[4];

   double vbar[4];

   long *ptr;

   long fffi[7];

   double fff[13];

   char iname[80];
   char fname[80];
   char bname[80];

   clock_t time1,timediff1;


   /*** Prints the usage. ***/


   if (argc < 2) {
      printf("\n");
      printf("Usage 1: %s algmodei relmodei barymodei printmodei\n",argv[0]);
      printf("                seedi ni tmaxr tpr dtr dtminr gr mfrr cr\n");
      printf("                rmaxr vmaxr mmaxr mminr outfile outfile.bin\n");
      printf("\n");
      printf("Usage 2: %s -file infile outfile outfile.bin\n",argv[0]);
      printf("\n");
      printf("                In file parameters:\n");
      printf("                algmodei relmodei barymodei printmodei\n");
      printf("                ni tmaxr tpr dtr dtminr gr mfrr cr\n");
      printf("                x(i) y(i) z(i) vx(i) vy(i) vz(i) m(i)\n");
      printf("\n");
      printf("Usage 3: %s -append outfile infile.bin tmaxr\n",argv[0]);
      printf("\n");
      printf("algmode: (0/4/5/8); BS 0; RK4 4; RK5 5; RK8 8.\n");
      printf("\n");
      printf("relmode: (0/1); Newtonian 0; Relativistic (PPN) 1.\n");
      printf("\n");
      printf("barymode: (0/1); Not subtract barycentrum 0; Subtract barycentrum 1.\n");
      printf("\n");
      printf("printmode: (0/1/2); Plot mode 0; Verify mode 1; Extended mode 2.\n");
      printf("\n");
      exit(0);
   } /* if */

   flag1=0;
   flag2=0;
   flag3=0;


   /*** Reads data for the -file mode from input file. ***/


if (argc == 5) {
   if (strcmp(argv[1],"-file")==0) {

      if ((infile = fopen(argv[2],"r")) == NULL) {
         printf("File not found.\n");
         exit(0);
      } /* if */

      strcpy(iname,argv[2]);
      strcpy(fname,argv[3]);
      strcpy(bname,argv[4]);

      flag1=1;
      flag2=1;
      inputflag=2;


#if LDOUBLE

      fscanf(infile,"%d",&algmode);
      fscanf(infile,"%d",&relmode);
      fscanf(infile,"%d",&barymode);
      fscanf(infile,"%d",&printmode);
      fscanf(infile,"%d",&n);
      fscanf(infile,"%Lf",&tmax);
      fscanf(infile,"%Lf",&tp);
      fscanf(infile,"%Lf",&ddt);
      fscanf(infile,"%Lf",&dtmin);
      fscanf(infile,"%Lf",&g);
      fscanf(infile,"%Lf",&mfr);
      fscanf(infile,"%Lf",&cc);

#else

      fscanf(infile,"%d",&algmode);
      fscanf(infile,"%d",&relmode);
      fscanf(infile,"%d",&barymode);
      fscanf(infile,"%d",&printmode);
      fscanf(infile,"%d",&n);
      fscanf(infile,"%lf",&tmax);
      fscanf(infile,"%lf",&tp);
      fscanf(infile,"%lf",&ddt);
      fscanf(infile,"%lf",&dtmin);
      fscanf(infile,"%lf",&g);
      fscanf(infile,"%lf",&mfr);
      fscanf(infile,"%lf",&cc);

#endif

      /*** Check for integration direction. ***/

      if (ddt < 0.0) {
         ddt=fabs(ddt);
	 dtsgn=-1.0;
      }
      else {
         ddt=fabs(ddt);
	 dtsgn=1.0;
      } /* if */
         

      /*** Adjusting stime-step for verify outout. ***/


      if (printmode == 1) {
         tp=tmax;
      } /* if */ 

      /*** Allocating primary arrays. ***/

      xx1=(double *)calloc(n+1,sizeof(double));
      yy1=(double *)calloc(n+1,sizeof(double));
      zz1=(double *)calloc(n+1,sizeof(double));
      vxx1=(double *)calloc(n+1,sizeof(double));
      vyy1=(double *)calloc(n+1,sizeof(double));
      vzz1=(double *)calloc(n+1,sizeof(double));

      mm0=(double *)calloc(n+1,sizeof(double));

      /*** Read body data. ***/

#if LDOUBLE

      for (j=1; j<=n; j++) {
          fscanf(infile,"%Lf %Lf %Lf %Lf %Lf %Lf %Lf",
	         &xx1[j],&yy1[j],&zz1[j],&vxx1[j],&vyy1[j],&vzz1[j],&mm0[j]);
      } /* for j */

#else

      for (j=1; j<=n; j++) {
          fscanf(infile,"%lf %lf %lf %lf %lf %lf %lf",
	         &xx1[j],&yy1[j],&zz1[j],&vxx1[j],&vyy1[j],&vzz1[j],&mm0[j]);
      } /* for j */

#endif


      fclose(infile);

   } /* if */
} /* if */


   /*** Reads data for the -append mode. ***/


if (argc == 5) {
   if (strcmp(argv[1],"-append")==0) {

      flag1=1;
      flag3=1;
      inputflag=3;

      if ((binfile = fopen(argv[3],"rb")) == NULL) {
         printf("File not found.\n");
         exit(0);
      } /* if */

      strcpy(fname,argv[2]);
      strcpy(bname,argv[3]);

      /*** Open and read binary data from binfile. ***/

      binfile=fopen(bname,"rb");

      fread(&fffi[1],sizeof(long),5,binfile);

      n=fffi[1];
      algmode=fffi[2];
      relmode=fffi[3];
      barymode=fffi[4];
      printmode=fffi[5];

      fread(&fff[1],sizeof(double),10,binfile);

      tmax0=fff[1];
      tp2=fff[2];
      ddt=fff[3];
      dt=fff[4];
      dtmin=fff[5];
      g=fff[6];
      mfr=fff[7];
      cc=fff[8];
      tt=fff[9];
      tp=fff[10];

      /*** Read tmax fron standard input. ***/

      tmax=atof(argv[4]);

      /*** Check for integration direction. ***/

      if (ddt < 0.0) {
         ddt=fabs(ddt);
	 dtsgn=-1.0;
      }
      else {
         ddt=fabs(ddt);
	 dtsgn=1.0;
      } /* if */

      /*** Adjusting stime-step for verify outout. ***/

      if (printmode == 1) {
         tp=tmax;
	 tp2=tp;
      } /* if */

      /*** Allocating primary arrays. ***/

      xx1=(double *)calloc(n+1,sizeof(double));
      yy1=(double *)calloc(n+1,sizeof(double));
      zz1=(double *)calloc(n+1,sizeof(double));
      vxx1=(double *)calloc(n+1,sizeof(double));
      vyy1=(double *)calloc(n+1,sizeof(double));
      vzz1=(double *)calloc(n+1,sizeof(double));

      mm0=(double *)calloc(n+1,sizeof(double));

      /*** Read body data from binary file. ***/

      for (j=1; j<=n; j++) {
         fread(&fff[1],sizeof(double),7,binfile);

	 xx1[j]=fff[1];
	 yy1[j]=fff[2];
	 zz1[j]=fff[3];
	 vxx1[j]=fff[4];
	 vyy1[j]=fff[5];
	 vzz1[j]=fff[6];
	 mm0[j]=fff[7];
         
     } /* for j */

     fclose(binfile);

   } /* if */
} /* if */


   if ((argc < 20) && (flag1 == 0)) {
      printf("\n");
      printf("Too few parameters !\n");
      printf("\n");
      exit(0);
   } /* if */


   /*** Reads data for generating spherical distribution of bodies ***/
   /*** from standard input.                                       ***/


   if (flag1 == 0) {

      flag2=1;
      inputflag=1;

      algmode=atoi(argv[1]);
      relmode=atoi(argv[2]);
      barymode=atoi(argv[3]);
      printmode=atoi(argv[4]);
      seed=atoi(argv[5]);
      n=atoi(argv[6]);
      tmax=atof(argv[7]);
      tp=atof(argv[8]);
      ddt=atof(argv[9]);
      dtmin=atof(argv[10]);
      g=atof(argv[11]);
      mfr=atof(argv[12]);
      cc=atof(argv[13]);
      rmax=atof(argv[14]);
      vmax=atof(argv[15]);
      mmax=atof(argv[16]);
      mmin=atof(argv[17]);
      strcpy(fname,argv[18]);
      strcpy(bname,argv[19]);

      /*** Check for integration direction. ***/

      if (ddt < 0.0) {
         ddt=fabs(ddt);
	 dtsgn=-1.0;
      }
      else {
         ddt=fabs(ddt);
	 dtsgn=1.0;
      } /* if */

      /*** Adjusting stime-step for verify outout. ***/


      if (printmode == 1) {
         tp=tmax;
      } /* if */


      /*** Allocating primary arrays. ***/

      xx1=(double *)calloc(n+1,sizeof(double));
      yy1=(double *)calloc(n+1,sizeof(double));
      zz1=(double *)calloc(n+1,sizeof(double));
      vxx1=(double *)calloc(n+1,sizeof(double));
      vyy1=(double *)calloc(n+1,sizeof(double));
      vzz1=(double *)calloc(n+1,sizeof(double));

      mm0=(double *)calloc(n+1,sizeof(double));

      idum=-abs(seed)-1;

      /*** Init random body data.  ***/

      init_random(mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,n,rmax,vmax,mmax,mmin,&idum);

   } /* if */


   /*** Checking for correct algmode, relmode and printmode. ***/

   if ((algmode != 0) && (algmode != 4) && (algmode != 5) && (algmode != 8)) {
      printf("\n");
      printf("Wrong algmode, choose (0/4/5/8) !\n");
      printf("\n");
      exit(0);
   } /* if */

   if ((relmode != 0) && (relmode != 1)) {
      printf("\n");
      printf("Wrong relmode, choose (0/1) !\n");
      printf("\n");
      exit(0);
   } /* if */

   if ((barymode != 0) && (barymode != 1)) {
      printf("\n");
      printf("Wrong barymode, choose (0/1) !\n");
      printf("\n");
      exit(0);
   } /* if */
      
   if ((printmode != 0) && (printmode != 1) && (printmode != 2)) {
      printf("\n");
      printf("Wrong printmode, choose (0/1/2) !\n");
      printf("\n");
      exit(0);
   } /* if */


   /*** Setting up global arrays. ***/

  axx1=(double *)calloc(n+1,sizeof(double));
  ayy1=(double *)calloc(n+1,sizeof(double));
  azz1=(double *)calloc(n+1,sizeof(double));

  qq=(double *)calloc(n+1,sizeof(double));

  if (relmode == 0) {

     axxn=(double *)calloc(1+1,sizeof(double));
     ayyn=(double *)calloc(1+1,sizeof(double));
     azzn=(double *)calloc(1+1,sizeof(double));

   }
   else {

     axxn=(double *)calloc(n+1,sizeof(double));
     ayyn=(double *)calloc(n+1,sizeof(double));
     azzn=(double *)calloc(n+1,sizeof(double));

   } /* if */

   if ((algmode == 5) || (algmode == 8)) {

      xerr1=(double *)calloc(n+1,sizeof(double));
      yerr1=(double *)calloc(n+1,sizeof(double));
      zerr1=(double *)calloc(n+1,sizeof(double));

   }
   else {

      xerr1=(double *)calloc(1+1,sizeof(double));
      yerr1=(double *)calloc(1+1,sizeof(double));
      zerr1=(double *)calloc(1+1,sizeof(double));

   } /* if */


   /*** Setting up arrays for Bulirsch-Stoer. ***/


  if (algmode == 0) {

     /*** Bulirsch-Stoer accuracy (bseps).                     ***/
     /*** If the routine bsstep exits without converging       ***/
     /*** make this constant larger (with say a factor of 10). ***/

#if LDOUBLE

     bseps=1.0e-17;

#else

     bseps=1.0e-14;

#endif

     n6=6*n;
     imax=11;

     xxx=(double *)calloc(imax+1,sizeof(double));

     yy=(double *)calloc(n6+1,sizeof(double));

     dydx=(double *)calloc(n6+1,sizeof(double));

     yscal=(double *)calloc(n6+1,sizeof(double));

     ddd=(double **)calloc(n6+1,sizeof(double*));
     mpddd=(double *)calloc((n6+1)*(imax+1),sizeof(double));

     for (i=1; i<=n6; i++) {
        ddd[i]=mpddd;
        mpddd=mpddd+imax+1;
     } /* for i */

     for (i=1; i<=n6; i++) {
       yscal[i]=1.0;
     } /* if */

  } /* if */


   /*** Setting up arrays for Runge-Kutta 8-order. ***/


  if (algmode == 8) {

     n6=6*n;
     nstage=13;

     yy=(double *)calloc(n6+1,sizeof(double));

     stages=(double **)calloc(n6+1,sizeof(double*));
     mpstages=(double *)calloc((n6+1)*(nstage+1),sizeof(double));

     for (i=0; i<=n6; i++) {
        stages[i]=mpstages;
        mpstages=mpstages+nstage+1;
     } /* for i */

     aaa=(double **)calloc(nstage+1,sizeof(double*));
     mpaaa=(double *)calloc((nstage+1)*(nstage+1),sizeof(double));

     for (i=0; i<=nstage; i++) {
        aaa[i]=mpaaa;
        mpaaa=mpaaa+nstage+1;
     } /* for i */


     bhat=(double *)calloc(nstage+1,sizeof(double));

     bbb=(double *)calloc(nstage+1,sizeof(double));

     ccc=(double *)calloc(nstage+1,sizeof(double));

     ptr=(long *)calloc(nstage+1,sizeof(long));


     rk8_init(ptr,aaa,bhat,bbb,ccc,nstage);

  } /* if */


/*** Printing mode information to standard output. ***/

printf("\n");
printf("Nbody\n");
printf("=====\n");
printf("\n");

if (inputflag == 1) {
   printf("Random spherical distribution of bodies.\n");
} /* if */

if (inputflag == 2) {
   printf("Input from text file:            %s\n",iname);
} /* if */

if (inputflag == 3) {
   printf("Input from binary file:          %s\n",bname);
} /* if */

printf("\n");

if (printmode == 0) {
   printf("Plot mode output to text file:   %s\n",fname);
   printf("\n");
   printf("Binary output to file:           %s\n",bname);
} /* if */

if (printmode == 1) {
   printf("Verify mode output to text file: %s\n",fname);
   printf("\n");
   printf("Binary output to file:           %s\n",bname);
} /* if */

if (printmode == 2) {
   printf("Extended mode output to text file:   %s\n",fname);
   printf("\n");
   printf("Binary output to file:               %s\n",bname);
} /* if */

printf("\n");

if (relmode != 1) {
   printf("Non relativistic (Newtonian) calculation.\n");
}
else {
   printf("Relativistic (PPN) calculation.\n");
} /* if */

printf("\n");

#if LDOUBLE

if (algmode == 0) {
printf("Bulirsch-Stoer integration (bseps = %9.3Le).\n",bseps);
} /* if */

#else

if (algmode == 0) {
printf("Bulirsch-Stoer integration (bseps = %9.3le).\n",bseps);
} /* if */

#endif

if (algmode == 4) {
printf("Runge-Kutta 4-order integration.\n");
} /* if */

if (algmode == 5) {
printf("Runge-Kutta 5-order integration.\n");
} /* if */

if (algmode == 8) {
printf("Runge-Kutta 8-order integration.\n");
} /* if */

printf("\n");

#if LDOUBLE

printf("Long double precision (80 bit numbers).\n");

#else

printf("Double precision (64 bit numbers).\n");

#endif

printf("\n");

if (barymode != 1) {
   printf("Not subtract barycentrum.\n");
}
else {
   printf("Subtract barycentrum.\n");
} /* if */

printf("\n");

printf("Number of bodies = %6d\n",n);

printf("\n");

#if LDOUBLE

printf("dtsgn = %2.0Lf\n\n",dtsgn);
printf("dt    = %18.12Lf; dtmin   = %18.12Lf\n\n",ddt,dtmin);
printf("Tmax  = %18.12Lf; Tmax/dt = %18.12Lf\n",tmax,tmax/ddt);

#else

printf("dtsgn = %2.0lf\n\n",dtsgn);
printf("dt    = %18.12lf; dtmin   = %18.12lf\n\n",ddt,dtmin);
printf("Tmax  = %18.12lf; Tmax/dt = %18.12lf\n",tmax,tmax/ddt);

#endif

printf("\n");

printf("Calculating...\n");

printf("\n");


/*** Setting initial position errors to zero. ***/


if ((algmode == 5) || (algmode == 8)) {

   for (j=1; j<=n; j++) {
      xerr1[j]=0.0;
      yerr1[j]=0.0;
      zerr1[j]=0.0;
   } /* for j */

} /* if */


   time1=clock();

/*** Calculate initial barycentrum position and velocity. ***/


   barycentrum(rbar,vbar,mm0,xx1,yy1,zz1,
               vxx1,vyy1,vzz1,n,g,mfr,cc,relmode);

   xbar=rbar[1];
   ybar=rbar[2];
   zbar=rbar[3];

   vxbar=vbar[1];
   vybar=vbar[2];
   vzbar=vbar[3];


/*** Subtract initial barycentrum data from body data. ***/

if ((barymode == 1) && (flag3 == 0)) {

   for (j=1; j<=n; j++) {

      xx1[j]=xx1[j]-xbar;
      yy1[j]=yy1[j]-ybar;
      zz1[j]=zz1[j]-zbar;

      vxx1[j]=vxx1[j]-vxbar;
      vyy1[j]=vyy1[j]-vybar;
      vzz1[j]=vzz1[j]-vzbar;

   } /* for j */

} /* if */


/*** Print initial values for the bodies (plot mode) to outfile. ***/


if (flag2 == 1) {

  outfile=fopen(fname,"w");

  if (printmode == 0) {

     fprintf(outfile,"%8d\n",n);

#if LDOUBLE

     for (j=1; j<=n; j++) { 
        fprintf(outfile,"%13.6Le %13.6Le %13.6Le %13.6Le\n",xx1[j],yy1[j],zz1[j],mm0[j]);
     } /* for j */

#else

     for (j=1; j<=n; j++) { 
        fprintf(outfile,"%13.6le %13.6le %13.6le %13.6le\n",xx1[j],yy1[j],zz1[j],mm0[j]);
     } /* for j */

#endif
 
  } /* if */

  if (printmode == 2) {

     fprintf(outfile,"%8d\n",n);

#if LDOUBLE

     for (j=1; j<=n; j++) { 
        fprintf(outfile,"%25.17Le %25.17Le %25.17Le %25.17Le %25.17Le %25.17Le %25.17Le\n",
                xx1[j],yy1[j],zz1[j],
                vxx1[j],vyy1[j],vzz1[j],mm0[j]);
     } /* for j */

#else

     for (j=1; j<=n; j++) { 
        fprintf(outfile,"%25.17le %25.17le %25.17le %25.17le %25.17le %25.17le %25.17le\n",
                xx1[j],yy1[j],zz1[j],
                vxx1[j],vyy1[j],vzz1[j],mm0[j]);
     } /* for j */

#endif

  } /* if */

} 
else {

  if (printmode != 1) {
     outfile=fopen(fname,"a");
  } 
  else {
     outfile=fopen(fname,"w");
  } /* if */

} /* if */

   
/*** Calculating first time-step. ***/

   
if (flag2 == 1) {

   compvmax(vxx1,vyy1,vzz1,n,&vvmax);

   comprmin(xx1,yy1,zz1,n,&rrmin);

   dt=ddt*rrmin/vvmax;

   if (dt > ddt) {
      dt=ddt;
   } /* if */

   if (dt < dtmin) {
      dt=dtmin;
   } /* if */

   tt=0.0;
   tp2=tp;

} /* if */

/*** Calculating some initial constants. ***/

   eps=0.1*dtmin;


/*** Starting main loop to tmax. ***/
   
  
while (tt < tmax) {

      h=dt*dtsgn;

      /*** Bulirsch-Stoer ***/


      if (algmode == 0) {

         bsstep(yy,dydx,h,bseps,yscal,ddd,xxx,imax,
                mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
                axxn,ayyn,azzn,qq,n,g,mfr,cc,&rrmin,relmode);


      } /* if */


      /*** Runge-Kutta 4-order. ***/


      if (algmode == 4) {

         rk4(h,
             mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
             axxn,ayyn,azzn,qq,n,g,mfr,cc,&rrmin,relmode);

      } /* if */


      /*** Runge-Kutta 5-order. ***/


      if (algmode == 5) {

         rk5(h,xerr1,yerr1,zerr1,
             mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
             axxn,ayyn,azzn,qq,n,g,mfr,cc,&rrmin,relmode);

      } /* if */


      /*** Runge-Kutta 8-order. ***/


      if (algmode == 8) {

         rk8(ptr,aaa,bhat,bbb,ccc,h,yy,stages,nstage,xerr1,yerr1,zerr1,
             mm0,xx1,yy1,zz1,vxx1,vyy1,vzz1,axx1,ayy1,azz1,
             axxn,ayyn,azzn,qq,n,g,mfr,cc,&rrmin,relmode);

      } /* if */



   /*** Print data to output file if print-step is reached. ***/

   tt=tt+dt;

   if (tt > (tp2-eps)) {
      for (j=1; j<=n; j++) { 

#if LDOUBLE

         if (printmode == 0) {
            fprintf(outfile,"%13.6Le %13.6Le %13.6Le %13.6Le\n",
	            xx1[j],yy1[j],zz1[j],mm0[j]);
         } /* if */

	 if ((printmode == 1) || (printmode == 2)) {
            fprintf(outfile,"%25.17Le %25.17Le %25.17Le %25.17Le %25.17Le %25.17Le %25.17Le\n",
	            xx1[j],yy1[j],zz1[j],
	            vxx1[j],vyy1[j],vzz1[j],mm0[j]);
	 } /* if */

#else
         if (printmode == 0) {
            fprintf(outfile,"%13.6le %13.6le %13.6le %13.6le\n",
	            xx1[j],yy1[j],zz1[j],mm0[j]);
         } /* if */

	 if ((printmode == 1) || (printmode == 2)) {
            fprintf(outfile,"%25.17le %25.17le %25.17le %25.17le %25.17le %25.17le %25.17le\n",
	            xx1[j],yy1[j],zz1[j],
	            vxx1[j],vyy1[j],vzz1[j],mm0[j]);
	 } /* if */

#endif



      } /* for j */
      tp2=tp2+tp;
   } /* if */


   /*** Calculate next time-step. ***/

   compvmax(vxx1,vyy1,vzz1,n,&vvmax);

   dt=ddt*rrmin/vvmax;

   if (dt > ddt) {
      dt=ddt;
   } /* if */

   if (dt < dtmin) {
      dt=dtmin;
   } /* if */

   if (dt > (tmax-tt)) {
      dt=tmax-tt;
   } /* if */


} /* while */



/*** Calculate final barycentrum position and velocity. ***/


barycentrum(rbar,vbar,mm0,xx1,yy1,zz1,
            vxx1,vyy1,vzz1,n,g,mfr,cc,relmode);


/*** Print Tend, CPU-time and barycentrum data to standard output. ***/

#if LDOUBLE

printf("Tend  = %18.12Lf\n",tt);
printf("\n");

timediff1=clock()-time1;
printf("CPU-time = %12.2Lf s\n",(double)(timediff1)/timetic);
printf("\n");

printf("                       X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");
printf("Rbar_ini = %22.12Le %22.12Le %22.12Le\n",xbar,ybar,zbar);
printf("\n");
printf("Rbar_end = %22.12Le %22.12Le %22.12Le\n",rbar[1],rbar[2],rbar[3]);
printf("\n");
printf("Vbar_ini = %22.12Le %22.12Le %22.12Le\n",vxbar,vybar,vzbar);
printf("\n");
printf("Vbar_end = %22.12Le %22.12Le %22.12Le\n",vbar[1],vbar[2],vbar[3]);


/*** For Runge-Kutta 5-order and 8-order print error between order 4 and 5 ***/
/*** and order 7 and 8 respectivley to standard output.                    ***/

if (algmode == 5) {

printf("\n");
printf("RK45-errors:           X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");

for (j=1; j<=n; j++) {
   printf("%10d %22.12Le %22.12Le %22.12Le\n",j,xerr1[j],yerr1[j],zerr1[j]);
} /* for j */ 

} /* if */

if (algmode == 8) {

printf("\n");
printf("RK78-errors:           X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");

for (j=1; j<=n; j++) {
   printf("%10d %22.12Le %22.12Le %22.12Le\n",j,xerr1[j],yerr1[j],zerr1[j]);
} /* for j */ 

} /* if */

printf("\n");

#else

printf("Tend  = %18.12lf\n",tt);
printf("\n");

timediff1=clock()-time1;
printf("CPU-time = %12.2lf s\n",(double)(timediff1)/timetic);
printf("\n");

printf("                       X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");
printf("Rbar_ini = %22.12le %22.12le %22.12le\n",xbar,ybar,zbar);
printf("\n");
printf("Rbar_end = %22.12le %22.12le %22.12le\n",rbar[1],rbar[2],rbar[3]);
printf("\n");
printf("Vbar_ini = %22.12le %22.12le %22.12le\n",vxbar,vybar,vzbar);
printf("\n");
printf("Vbar_end = %22.12le %22.12le %22.12le\n",vbar[1],vbar[2],vbar[3]);


/*** For Runge-Kutta 5-order and 8-order print error between order 4 and 5 ***/
/*** and order 7 and 8 respectivley to standard output.                    ***/

if (algmode == 5) {

printf("\n");
printf("RK45-errors:           X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");

for (j=1; j<=n; j++) {
   printf("%10d %22.12le %22.12le %22.12le\n",j,xerr1[j],yerr1[j],zerr1[j]);
} /* for j */ 

} /* if */

if (algmode == 8) {

printf("\n");
printf("RK78-errors:           X                      Y                      Z\n");
printf("-------------------------------------------------------------------------------\n");

for (j=1; j<=n; j++) {
   printf("%10d %22.12le %22.12le %22.12le\n",j,xerr1[j],yerr1[j],zerr1[j]);
} /* for j */ 

} /* if */

printf("\n");

#endif

fclose(outfile);


/*** Store final data in binary file. ***/


binfile=fopen(bname,"wb");

fffi[1]=n;
fffi[2]=algmode;
fffi[3]=relmode;
fffi[4]=barymode;
fffi[5]=printmode;

fwrite(&fffi[1],sizeof(long),5,binfile);

fff[1]=tmax;
fff[2]=tp2;
fff[3]=ddt*dtsgn;
fff[4]=dt;
fff[5]=dtmin;
fff[6]=g;
fff[7]=mfr;
fff[8]=cc;
fff[9]=tt;
fff[10]=tp;

fwrite(&fff[1],sizeof(double),10,binfile);

for (j=1; j<=n; j++) {
   fff[1]=xx1[j];
   fff[2]=yy1[j];
   fff[3]=zz1[j];
   fff[4]=vxx1[j];
   fff[5]=vyy1[j];
   fff[6]=vzz1[j];
   fff[7]=mm0[j];
   
   fwrite(&fff[1],sizeof(double),7,binfile);

} /* for j */

fclose(binfile);


} /* End */

