/***********************************************************************/
/*                                                                     */
/*                                                                     */
/* Program:                                                            */
/* --------                                                            */
/*                                                                     */
/* pparam.c                                                            */
/*                                                                     */
/*                                                                     */
/* Version:                                                            */
/* --------                                                            */
/*                                                                     */
/* First version: February 2000                                        */
/* Latest update: 2003-02-18                                           */
/*                                                                     */
/*                                                                     */
/* Author of the program:                                              */
/* ----------------------                                              */
/*                                                                     */
/* Stefan Spaennare, Lund, Sweden                                      */
/*                                                                     */
/* E-mail: stefan@spaennare.se                                         */
/*                                                                     */
/* Web: http://www.spaennare.se/index.html                             */
/* Web: http://www.spaennare.se/ssnbody.html                           */
/*                                                                     */
/* If you have problems with the program please let me know!           */
/*                                                                     */
/*                                                                     */
/* Documentation:                                                      */
/* --------------                                                      */
/*                                                                     */
/* For a detailed description of the ssnbody nbody package see the     */
/* files ssnbody.pdf and ssnbody.ps on the ssnbody web page.           */
/*                                                                     */
/*                                                                     */
/* References:                                                         */
/* -----------                                                         */
/*                                                                     */
/* This program was written with permission from Staffan Soederhjelm,  */
/* Lund Observatory. The C-program was translated by the author from   */
/* a Fortran routine written by him.                                   */
/*                                                                     */
/***********************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

double pi=3.14159265358979323;


/* Converts radians to degrees. */

double deg(x)
double x;
{
   return(180.0*x/pi);
} /* deg */


/* Converts degrees to radians. */

double rad(x)
double x;
{
  return(pi*x/180.0);
} /* rad */


/*******************************************************************/
/*                                                                 */
/* Subroutine to calculate osculating elements from r,rdot vectors */
/* (for simplicity translated Montenbruck Pascal-routine)          */
/*                                                                 */
/* Input: gm           effective GM                                */
/*        elin(1-3)    x,y,z  (au)                                 */
/*        elin(4-6)    xdot,ydot,zdot  (au/day)                    */
/* Output: elout(1)    a   (au)                                    */
/*         elout(2)    e                                           */
/*         elout(3)    incl (rad)                                  */
/*         elout(4)    lom  (rad)  !from 9911 lom+stom             */
/*         elout(5)    stom (rad)                                  */
/*         elout(6)    M (mean anomaly, rad)                       */
/*                                                                 */
/*  SSo/960401, lom=long instead of arg,0-tpi constraint 9911      */
/*                                                                 */
/*******************************************************************/

void oscel(gm,elin,elout)
double gm;
double *elin;
double *elout;
{

   double hx,hy,hz,h;
   double u,r,v2,ax,c,s,e2,nu,ea,tpi;

   tpi=2.0*pi;

   /* Areal velocity => inc, stom */

   hx=elin[2]*elin[6]-elin[3]*elin[5];
   hy=elin[3]*elin[4]-elin[1]*elin[6];
   hz=elin[1]*elin[5]-elin[2]*elin[4];
   h=sqrt(hx*hx+hy*hy+hz*hz);

   elout[5]=atan2(hx,-hy);                    /* stom             */
   elout[3]=atan2(sqrt(hx*hx+hy*hy),hz);      /* incl             */
   u=atan2(elin[3]*h,-elin[1]*hy+elin[2]*hx); /* Arg. of latitude */

   /* Energy => a */

   r=sqrt(elin[1]*elin[1]+elin[2]*elin[2]+elin[3]*elin[3]);
   v2=elin[4]*elin[4]+elin[5]*elin[5]+elin[6]*elin[6];
   ax=1.0/(2.0/r-v2/gm);
   elout[1]=ax; /* a */

   /* ecc, E */

   c=1.0-r/ax; /* e*cos(E) */
   s=(elin[1]*elin[4]+elin[2]*elin[5]+elin[3]*elin[6])/sqrt(ax*gm); /* e*sin(E) */
   e2=c*c+s*s;

   elout[2]=sqrt(e2);             /* e                 */
   ea=atan2(s,c);                 /* Eccentric anomaly */
   elout[6]=ea-s;                 /* Mean anomaly      */
   nu=atan2(sqrt(1.0-e2)*s,c-e2); /* True anomaly      */
   elout[4]=u-nu;                 /* lom               */

   elout[4]=elout[4]+elout[5];

   if (elout[4] < 0.0) {
      elout[4]=elout[4]+tpi;
   } /* if */

   if (elout[4] > tpi) {
      elout[4]=elout[4]-tpi;
   } /* if */

   if (elout[6] < 0.0) {
      elout[6]=elout[6]+tpi;
   } /* if */

   if (elout[6] > tpi) {
      elout[6]=elout[6]-tpi;
   } /* if */

} /* oscel */


int main(argc,argv) 
int argc;
char *argv[];

{
   
   FILE *infile;

   int i,temp;


   double x,y,z,vx,vy,vz,m;
   double x2,y2,z2,vx2,vy2,vz2,m2;
   double g,msol,au,a,p,mm,e;
   double sday,mmsol,ftemp;

   double elin[7];
   double elout[7];

   if (argc != 2) {
      printf("Usage: %s pulsar-infile\n",argv[0]);
      exit(0);
   } /* if */
   

   if ((infile = fopen(argv[1],"r")) == NULL) {
      printf("File not found.\n");
      exit(0);
   } /* if */


   sday=3600.0*24.0;

   au=1.49597870684e11;

   g=6.670e-11;

   mmsol=2.959122082856e-4;

   msol=1.989e30;


   for (i=1; i<=5; i++) {
      fscanf(infile,"%d",&temp);
   } /* for i */

   for (i=1; i<=7; i++) {
      fscanf(infile,"%lf",&ftemp);
   } /* for i */

   fscanf(infile,"%lf %lf %lf %lf %lf %lf %lf",&x,&y,&z,&vx,&vy,&vz,&m);
   fscanf(infile,"%lf %lf %lf %lf %lf %lf %lf",&x2,&y2,&z2,&vx2,&vy2,&vz2,&m2);

   elin[1]=x-x2;
   elin[2]=y-y2;
   elin[3]=z-z2;

   elin[4]=vx-vx2;
   elin[5]=vy-vy2;
   elin[6]=vz-vz2;

   mm=m+m2;

   oscel(mm,elin,elout);

   a=elout[1];

   e=elout[2];

   p=sqrt((4.0*pi*pi*a*a*a)/mm);

   printf("\n");

   printf("Pulsar data\n");
   printf("===========\n");

   printf("\n");
   
   printf("m1 = %13.6lf Msun\n",m/mmsol);

   printf("\n");

   printf("m2 = %13.6lf Msun\n",m2/mmsol);

   printf("\n");

   printf("a =  %13.6le m = %12.8lf au\n",au*a,a);

   printf("\n");

   printf("P =  %13.2lf s = %12.8lf days\n",sday*p,p);
   
   printf("\n");

   printf("e =  %13.6lf\n",e);

   printf("\n");

} /* End */
